/************************************************/
/* file_system.c                                */
/* Mu-huan Chiang, Brandon Ellison              */
/*                                              */
/************************************************/

#ifndef M16C_62P
#include <stdio.h>
#include <string.h>
#endif

#include "file_system.h"

/* Allocate the space for the files */
file_entry FAT[MAX_FILES] = { {1,"/",0,0,{0,0,0,0,0},0},};
block RAM[MAX_BLOCK_COUNT];

UINT8 num_files = 1;
UINT8 block_count = 0;

//Internal functions only
	//Removes a file entry from its parent directory
void RemoveFileFromDirectory(UINT8 fat_index);	
	//Finds an available block for use or 0 if there is none.
block * GetEmptyBlock();
void our_strcpy(far char * dest, far char * src);
void our_strcat(far char * orig, far char * append);
int our_strcmp(far char * str1, far char * str2);


//FindFile
//
// This function finds the index of the file in the FAT table, this requires
// a pathname and a filename.
// Returns the index of the given file.

INT8 FindFile(UINT32 name_str)
{
	UINT8 i;
	char * str1, * str2;

	//Loop through the FAT table looking at each entry
	// for naming information.
	for(i = 0; i < MAX_FILES; i++)
	{
		if (FAT[i].type == 0)
			continue;		

		// Compare the strings 
		str1 = (char *) name_str;
		str2 = (char *) FAT[i].name;
		*str2++;

		//Look through the strings for exact match.
		while(*str1 == *str2)
		{
			// if both at the null, then strings match
			if( *str1 == 0)
				return i;
			else
			{
				str1++;
				str2++;
			}
		}
	}

	// not found, return -1
	return -1;
}


//FindFilePath
//
// This function finds the index of the file in the FAT table, this requires
// a pathname and a filename.
// Returns the index of the given file.

INT8 FindFilePath(UINT32 path, UINT32 name_str)
{
	UINT8 i;
	char * str1, * str2;
	char strbuffer[50] = {0,};

	//If the given path is not the root path, copy it to the string buffer.
	if (our_strcmp("/", (far char *)path) != 0)
		our_strcat(strbuffer, (far char *)path);

	//If the name is not empty, concatenate a / to the buffer.
	if (our_strcmp("", (far char *)name_str) != 0)
		our_strcat(strbuffer, "/");

	//If the path is / and the name is blank, then append a / to the buffer.
	if (our_strcmp("/", (far char *)path) == 0 && our_strcmp("",(far char *)name_str) == 0)
		our_strcat(strbuffer, "/");

	//Add the name to the sting buffer, completing the pathname.
	our_strcat(strbuffer, (far char *)name_str);

	//Loop through the FAT table looking at each entry
	// for naming information.
	for(i = 0; i < MAX_FILES; i++)
	{
		if (FAT[i].type == 0)
			continue;		

		// Compare the strings 
		str1 = (char *) strbuffer;
		str2 = (char *) FAT[i].name;

		//Look through the strings for exact match.
		while(*str1 == *str2)
		{
			// if both at the null, then strings match
			if( *str1 == 0)
				return i;
			else
			{
				str1++;
				str2++;
			}
		}
	}

	// not found, return -1
	return -1;
}

//GetFileSize
//
// This returns the file size of the given file index.

UINT16 GetFileSize( UINT8 fat_index )
{
	return FAT[fat_index].size;
}


//ReadFile
// This takes the given file index, and reads the designated
// number of bytes and returns it.

UINT16 ReadFile(UINT8 fat_index, UINT16 offset, UINT8 * buffer, UINT16 bytes2read)
{
	UINT16 i, block_end;
	UINT8 copy_from;
	UINT16 bytes_read = 0;

	//Pointer to current block to start reading from
	block * c_block;

	// If bytes2read exceeds the file size, read till the end of file
	if (bytes2read + offset > FAT[fat_index].size) 
		bytes2read = FAT[fat_index].size - offset;

	// First find the block in which the offset starts in, then copy
	// out the data. If we are at the end of a block and we still
	// need more data, then start reading from the next block.
	if ( bytes2read > 0)
	{
		// Find the block begining with the offset
		block_end = 0;
		copy_from = 0;
		c_block = FAT[fat_index].start_block;
		block_end += BLOCK_SIZE;

		//Loop until we find the block to start reading from
		while (block_end <= offset)
		{
			//If we're reading past the end of the file, return an error.
			if (c_block->next_block ==  0)
				return 0;

			//Move to the next block and increment
			c_block = c_block->next_block;
			block_end += BLOCK_SIZE;
			if (offset < block_end)
				break;		// found the block to start reading from
		}


		// Get the offset (in the block) from where we will start reading from
		// this will give us the offset from the begining of the block
		copy_from = offset % BLOCK_SIZE;		

		// Start copying
		while (bytes2read-bytes_read > 0)
		{
			//Move data, byte by byte, but first check to see
			// that we are not too far into block.
			if (copy_from == BLOCK_SIZE)
			{
				//If we are past the end of the block, move
				// to the next block if there is one and continue.
				if (c_block->next_block != 0)
				{
					c_block = c_block->next_block;
					copy_from = 0;
				} 
				else
				{
					//We've hit end, call it quits.
					bytes2read = 0;
					break;
				}
			}
			//Move a byte of data to the buffer.
			*(UINT8 *)buffer++ = c_block->block_data[copy_from++];
			bytes_read++;
		}
	}
		
	return bytes_read;
}

//CreateDirectory
//
// This creates a directory.  It needs a base path to start from and a directory
// name to use.

INT8 CreateDirectory(UINT32 path, UINT32 dirname)
{
	int i;
	UINT8 index;
	INT8 path_index;
	char strbuffer[50] = {0,};

	//Search for the first unused entry.
	for (i = 0; i < MAX_FILES; i++)
	{
		//Claim it as ours.
		if (FAT[i].type == 0)
		{
			index = i;
			break;
		}
	}

	//Find the parent directory
	path_index = FindFilePath(path,(UINT32)"");

	//Make sure the file system can handle the addition.
	if ((num_files < MAX_FILES) && (path_index != -1) && 
		(FAT[path_index].num_dirfiles < MAX_DIRFILES))
	{
		//Build a new path name. If the path is not the
		// root path start the buffer with it.
		if (our_strcmp("/", (far char *)path) != 0)
		{
			our_strcat(strbuffer, (far char *)path);
		}

		//Add the / and the filename as part of the path
		our_strcat(strbuffer, "/");
		our_strcat(strbuffer, (far char *) dirname);

		//Set the parameters for our new directory
		// This is basic initialization for the directory node.
		FAT[index].type = 1;
		our_strcpy(FAT[index].name, (far char *) strbuffer);
		FAT[index].size = 0;
		FAT[index].num_dirfiles = 0;
		for (i = 0; i < MAX_DIRFILES; i++)
		{
			FAT[index].dirfiles[i] = -1;
		}
		FAT[index].start_block = 0;

		//Setup the parent directory to include this new directory.
		FAT[path_index].dirfiles[FAT[path_index].num_dirfiles] = index;
		FAT[path_index].num_dirfiles++;

		//Increment the number of files and return the new index.
		num_files++;
		return index;
	}

	//Something failed out, return error.
	return -1;	
}

//CreateFile
//
// This creates a new file entry, it needs a base path and a filename.
// This returns the index to the file.

INT8 CreateFile(UINT32 path, UINT32 name_str) 
{
	int i;
	UINT8 index;
	INT8 path_index;
	char strbuffer[50] = {0,};

	//Search for the first unused entry.
	for (i = 0; i < MAX_FILES; i++)
	{
		//Claim it as ours.
		if (FAT[i].type == 0)
		{
			index = i;
			break;
		}
	}

	//Find the parent directory
	path_index = FindFilePath(path,(UINT32)"");

	//Make sure the file system can handle the addition.
	if ((num_files < MAX_FILES) && (path_index != -1) &&
		(FAT[path_index].num_dirfiles < MAX_DIRFILES))
	{
		//Build a new path name. If the path is not the
		// root path start the buffer with it.
		if (our_strcmp("/",(far char *)path) != 0)
			our_strcat(strbuffer, (far char *)path);

		//Add the / and the filename as part of the path
		our_strcat(strbuffer, "/");
		our_strcat(strbuffer, (far char *)name_str);

		//Set the parameters for our new file
		// This is basic initialization for the file node.
		FAT[index].type = 2;
		our_strcpy(FAT[index].name,(far char *)strbuffer);
		FAT[index].size = 0;
		FAT[index].num_dirfiles = 0;
		for (i = 0; i < MAX_DIRFILES; i++)
			FAT[index].dirfiles[i] = -1;
		FAT[index].start_block = 0;

		//Setup the parent directory to include this new directory.
		FAT[path_index].dirfiles[FAT[path_index].num_dirfiles] = index;
		FAT[path_index].num_dirfiles++;

		//Increment the number of files and return the new index.
		num_files++;
		return index;
	}

	//Something failed out, return error.
	return -1;	
}

//DeleteFile
//
// This deletes a file from the system and returns the blocks to the system.
// This returns the index to the file.

INT8 DeleteFile(UINT8 fat_index)
{
	int i,j,k;
	block * p_block;

	//Find the starting block.
	block * c_block = FAT[fat_index].start_block;

	//Loop until all blocks are empty.
	while (c_block != 0)
	{
		//Find the last block in the chain.
		p_block = 0;
		while (c_block->next_block != 0)
		{
			p_block = c_block;
			c_block = c_block->next_block;
		}

		//Mark the last block as unused and then mark
		// the parent block's next pointer to null.
		c_block->used = 0;
		c_block->next_block = 0;
		block_count--;

		if (p_block != 0)
			p_block->next_block = 0;
		else
		{
			//We are the last block in, so clear out the file.
			FAT[fat_index].start_block = 0;
		}
		//Reset the start block and go around again.
		c_block = FAT[fat_index].start_block;
	}
	//Mark the block as unused.
	FAT[fat_index].type = 0;

	//Clear out the entry from the directory listing.
	RemoveFileFromDirectory(fat_index);

	//Bookkeeping.
	num_files--;

	//Return the entry of the file deleted.
	return fat_index;
}

//RemoveDirectory
//This function removes a directory from the FAT table.  All that
// is needed is the path of the directory.
INT8 RemoveDirectory(UINT32 dirpath) 
{
	int i,j,k;
	INT8 index;
	
	//Find the directory in the FAT table.
	index = FindFilePath(dirpath,(UINT32)"");
	
	//If the directory doesn't exist, of if is not a directory entry, or if
	// the directory is not empty, error out.
	if (index == -1 || FAT[index].type != 1 || FAT[index].num_dirfiles > 0)
	{
		return -1;
	}

	//Remove this listing from the directory tables.
	RemoveFileFromDirectory(index);

	//Return the index of the deleted record.
	FAT[index].type = 0;
	return index;
}

//WriteFile
//This function adds data to a file, either overwriting old data
// or extending the file size to compensate.  This function takes in an offset
// and a # of bytes to write, as well as the byte buffer.

UINT16 WriteFile(UINT8 fat_index, UINT16 offset, UINT8 * buffer, UINT16 bytes2write)
{
	//Setup.
	UINT16 block_end;
	block* c_block;
	block** s_block;
	UINT8 copy_to;
	UINT16 bytes_written = 0;

	// If it's a new file, create the start_block 
	s_block = &(FAT[fat_index].start_block);
	c_block = FAT[fat_index].start_block;
	if (c_block == 0)
	{
		//First block is null, so we have no data in the file yet.
		if (block_count < MAX_BLOCK_COUNT)
		{
			//Grab an empty block.
			*s_block = GetEmptyBlock();
		}
		else 
		{
			//Not enough empty blocks, call it out.
#ifndef M16C_62P
			printf("WriteFile Error: Not enough free blocks!!\n");
#endif
			return 0;
		}
	}

	//Check if the offset is NOT larger than the file size
	// to extend a file, use the current file size as the offset.
	if (offset > FAT[fat_index].size) 
	{
#ifndef M16C_62P
		printf("WriteFile Error: Offset is larger than file size!!\n");
#endif
		return 0;
	}

	//Find the block begining with the offset
	//Start at the beginning of the file and move ahead
	block_end = 0;
	c_block = FAT[fat_index].start_block;
	block_end += BLOCK_SIZE;
	while (offset > block_end)
	{
		//Did we reach the end without finding the offset.
		if (c_block->next_block == 0)
		{
#ifndef M16C_62P
			printf("WriteFile Error: Offset is larger than file size %i %i %i!\n",
				offset, block_end, FAT[fat_index].size);
#endif
			return 0;
		}

		//Move forward until the block is found
		c_block = c_block->next_block;
		block_end += BLOCK_SIZE;
		if (offset <= block_end )
			break;		// found the block to start writing to
	}
	
	//Get the offset (in the block) from where we will start reading from
	// this will give us the offset from the begining of the block
	if (offset == block_end)
		copy_to = BLOCK_SIZE;
	else
		copy_to = offset % BLOCK_SIZE;		

	//Start copying data from the input buffer
	while (bytes2write - bytes_written > 0)
	{
		//Move data, byte by byte, but first check to see
		// that we are not too far into block.
		if (copy_to == BLOCK_SIZE)
		{
			//If we are past the end of the block, move
			// to the next block if there is one and continue.
			if (c_block->next_block != 0)
			{
				c_block = c_block->next_block;
				copy_to = 0;
			} 
			else 
			{
				//There are available blocks left in the system.
				// Grab an empty block, and use it.  Append it to the
				// file.
				if (block_count < MAX_BLOCK_COUNT)
				{
					c_block->next_block = GetEmptyBlock();
					c_block = c_block->next_block;
					copy_to = 0;
				} 
				else
				{
					//File system is full, call it quits.
#ifndef M16C_62P
					printf("Not enough free blocks!!\n");
#endif
					return bytes_written;
				}
			}
		}
		//Move a byte of data to the buffer.
		c_block->block_data[copy_to++] = *(UINT8 *)buffer++;
		bytes_written++;
	}

	//Add filesize to the file as needed.
	FAT[fat_index].size = offset + bytes_written;

	//Return the bytes written.
	return bytes_written;
}

//GetEmptyBlock
//
//This function finds an empty block in the system and returns it.

block* GetEmptyBlock()
{
	UINT8 i;
	block* e_block;

	//Search all the blocks
	for (i = 0; i < MAX_BLOCK_COUNT; i++)
	{
		//If an unused block is found, claim it as used and return it.
		if (RAM[i].used == 0)
		{
			RAM[i].next_block = 0;
			e_block = &(RAM[i]);
			e_block->used = 1;
			block_count++;
			return e_block;
		}
	}
	return 0;
}

//MoveFile
//
//This function moves a file from one directory to another.

INT8 MoveFile(UINT32 dir1, UINT32 file1, UINT32 dir2, UINT32 file2)
{
	int i;
	INT8 i1, i2, i3;
	char strbuffer[50] = {0,};

	//Find both files in question.
	i1 = FindFilePath(dir1, file1);
	i2 = FindFilePath(dir2, file2);
	i3 = FindFilePath(dir2, (UINT32)"");

	//If the first file can't be found, abort.
	if (i1 == -1)
		return -1;

	//If the path of the second file can't be found, abort.
	if (i3 == -1)
		return -1;

	//If the second file exists, delete it.
	if (i2 != -1)
		DeleteFile(i2);

	//Make sure the file system can handle the addition.
	// Do not want to overload the directory structure.
	if (FAT[i3].num_dirfiles == MAX_DIRFILES)
		return -1;

	//Create a new file name from the given pieces.
	if (our_strcmp("/",(far char *)dir2) != 0)
		our_strcat(strbuffer, (far char *)dir2);
	our_strcat(strbuffer, "/");
	our_strcat(strbuffer, (far char *)file2);

	RemoveFileFromDirectory(i1);

	//Setup the parent directory to include this new directory.
	FAT[i3].dirfiles[FAT[i3].num_dirfiles] = i1;
	FAT[i3].num_dirfiles++;

	//Clear the old filename
	for (i = 0; i < MAX_FILENAME + 1; i++)
		FAT[i1].name[i] = 0;	

	//Copy the filename over, and return.
	our_strcpy(FAT[i1].name, (far char *)strbuffer);
	return i1;
}

//IsDirectory
//
//This function checks the filepath to see if the given 
// directory name exists.

INT8 IsDirectory(UINT32 filepath)
{
	INT8 fat_index;

	//Find the path in the system
	fat_index = FindFilePath(filepath, (UINT32)"");
	if (fat_index == -1)
		return 0;
	
	//If it is a directory entry, return true, else false.
	if (FAT[fat_index].type == 1)
		return 1;
	return 0;
}

//Format
//
//This function re-initializes the filesystem to start over from
// scratch. Returns 0 on success.

INT8 Format()
{
	int i;
	
	//Mark all blocks as unused.
	for (i = 0; i < MAX_BLOCK_COUNT; i++)
	{
		RAM[i].used = 0;
		RAM[i].next_block = 0;
	}

	//Mark all file entries as empty.
	for (i = 0; i < MAX_FILES; i++)
	{
		FAT[i].type = 0;
		our_strcpy(FAT[i].name, "");
		FAT[i].size = 0;
		FAT[i].num_dirfiles = 0;
		FAT[i].start_block = 0;
	}

	//Setup initial directory.
	FAT[0].type = 1;
	our_strcpy(FAT[0].name, "/");

	//All blocks are empty, only one file in use.
	block_count = 0;
	num_files = 1;

	return 0;
}

//List
//
//This function returns the files in a directory.

INT8* List(UINT32 dirpath)
{
	INT8 dir_index;

	//Find the path.
	dir_index = FindFilePath(dirpath,(UINT32)"");
	if (dir_index == -1 || FAT[dir_index].type != 1)
		return 0;
	
	//Return the directory file entries.
	return FAT[dir_index].dirfiles;
}

//SpaceStatus
//
//This function is for debug and prints the statistics of the file system.

void SpaceStatus(UINT16 * used_files, UINT16 * used_blocks, UINT16 * used_space, UINT16 * bytes_total)
{
	UINT16 max_files, max_blocks, max_space;
	UINT16 directories, i;

	max_files = 0; *used_files = 0;
	max_blocks = 0; *used_blocks = 0;
	max_space = 0; *used_space = 0;
	directories = 0; *bytes_total = 0;
	for (i = 0; i < MAX_FILES; i++)
	{
		max_files++;
		if (FAT[i].type == 1)
		{
			(*used_files)++;
			directories++;
		}
		if (FAT[i].type == 2)
		{
			(*used_files)++;
			(*used_space) += FAT[i].size;
		}
	}

	for (i = 0; i < MAX_BLOCK_COUNT; i++)
	{
		max_blocks++;
		if (RAM[i].used)
			(*used_blocks)++;
		max_space += BLOCK_SIZE;
	}

	*bytes_total = sizeof(file_entry) * MAX_FILES;
	*bytes_total += sizeof(block) * MAX_BLOCK_COUNT;
	*bytes_total += 2 * sizeof(UINT8);

#ifndef M16C_62P
	printf("System: %u files used of (%i max) (%i are dirs), %u blocks used (%i max)\n", 
		*used_files, max_files, directories, *used_blocks, max_blocks);
	printf("  %u bytes used (%i max). %u total RAM used\n", 
		*used_space, max_space, *bytes_total);
#endif
}


//DirListing
//
//This function is for debug and prints the directory listings.

void DirListing(UINT8 dir)
{
	UINT8 i, printed;

#ifndef M16C_62P
	printf("Directory listing for: %s\n", FAT[dir].name);
	printed = 0;
	for (i = 0; i < FAT[dir].num_dirfiles; i++)
	{
		if ((FAT[dir].dirfiles[i] != -1) && 
			(FAT[FAT[dir].dirfiles[i]].type == 2))
		{
			if (printed == 0)
				printf("\t"); 
			printf("%s ", FAT[FAT[dir].dirfiles[i]].name);
			printed++;
		}
	}
	
	if (printed)
		printf("\n");

	for (i = 0; i < FAT[dir].num_dirfiles; i++)
	{
		if ((FAT[dir].dirfiles[i] != -1) && 
			(FAT[FAT[dir].dirfiles[i]].type == 1))
		{
			DirListing(FAT[dir].dirfiles[i]);
		}
	}
#endif
}

//RemoveFileFromDirectory
//
//This function removes an entry from a directory.
//This saves duplication of code.

void RemoveFileFromDirectory(UINT8 fat_index)
{
	UINT8 i, j, k;

	//Find the file in the listings.
	for (i = 0; i < MAX_FILES; i++)
	{
		//Check all directories.
		if (FAT[i].type == 1)
		{
			//If it's a directory, check the entries.
			for (j = 0; j < FAT[i].num_dirfiles; j++)
			{
				//If this is the entry...
				if (FAT[i].dirfiles[j] == fat_index)
				{
					//Rewrite the directory to remove the file.
					for (k = j; k < MAX_DIRFILES - 1; k++)
					{
						FAT[i].dirfiles[k] = FAT[i].dirfiles[k + 1];
					}
					//Update the last entry and clear out.
					FAT[i].dirfiles[MAX_DIRFILES - 1] = -1;
					FAT[i].num_dirfiles--;
					break;
				}
			}
		}
	}
}

//DumpFile
//
//This function is for debug and dumps a file to the OS.

void DumpFile(UINT8 index, UINT8 * name)
{
#ifndef M16C_62P
	UINT8 buffer[5];
	int i, num_bytes;
	FILE * file = fopen((char *)name, "w");
	for (i = 0; i < FAT[index].size; i++)
	{
		num_bytes = ReadFile(index, i, buffer, 1);
		fputc(buffer[0], file);
	}
	fclose(file);
#endif
}


int our_strcmp(far char * str1, far char * str2)
{
	int index = 0;
	while (str1[index] == str2[index])
	{
		if (str1[index] == 0)
			return 0;
		index++;
	}
	return 1;
}

void our_strcat(far char * orig, far char * append)
{
	int index1 = 0;
	int index2 = 0;
	while (orig[index1] != 0)
		index1++;

	while (append[index2] != 0)
	{
		orig[index1] = append[index2];
		index1++;
		index2++;
	}

	orig[index1 + 1] = 0;
}

void our_strcpy(far char * dest, far char * src)
{
	int index1 = 0;
	while (src[index1] != 0)
	{
		dest[index1] = src[index1];
		index1++;
	}
	dest[index1] = 0;
}
